#ifndef CUDA_TIMER_H
#define CUDA_TIMER_H

#include <cuda.h>
#include <cuda_runtime.h>
#include <device_launch_parameters.h>

/* 
 * Tiny Class to measure elapsed time between cuda commands
 * usage:
 *  CudaTimer timer; // get an instance of the class
 *  timer.start(); // starts timer
 *  // your cuda code
 *  timenElapsed = timer.elapsed(); // stops timer and returns elapsed time in milliseconds
 */
class CudaTimer
{
public:
	CudaTimer();
	~CudaTimer();
	void start( void );
	float elapsed( void ) const;			//!< time in milliseconds

private:
	cudaEvent_t		m_start;
	cudaEvent_t		m_stop;
};

inline CudaTimer::CudaTimer()
{
	cudaEventCreate(&m_start);
	cudaEventCreate(&m_stop);
}

inline CudaTimer::~CudaTimer()
{
	cudaEventDestroy(m_start);
	cudaEventDestroy(m_stop);
}

inline void CudaTimer::start( void )
{
	cudaEventRecord(m_start);
	cudaEventSynchronize(m_start);
}

inline float CudaTimer::elapsed( void ) const
{
	cudaEventRecord(m_stop);
	cudaEventSynchronize(m_stop);

	float elapsedTime;
	cudaError_t err = cudaEventElapsedTime(&elapsedTime, m_start, m_stop);
	if ( cudaSuccess != err )
		elapsedTime = -1.0f;

	return elapsedTime;
}

#endif //CUDA_TIMER_H