/****************************************************************************************************\
 *  GPGPU exercise:                                                                                 *
 *  (simplified) simulation of heat transfer                                                        *
\****************************************************************************************************/

// include some convenience functions (names begin with "ue_" to distinguish them from CUDA)
#include "utility_environment.h"

// OpenGL utilities
#include "opengl_includes.h"

#include <vector>
#include <cstdlib>
#include <algorithm>

////////////////////////////////////////////////////////////////////////////////
// GLUT, which we use for displaying images, makes use of callback functions. //
// The following global variables are the simplest way of passing parameters  //
// between the various callback functions.                                    //
////////////////////////////////////////////////////////////////////////////////

int g_imgWidth = 2*512;                   // size of the simulation array and the displayed image
int g_imgHeight = 2*512;

float g_mu = 0.1f;

float *h_Buffer = 0;                    // host buffer for copy of GPU results
float *d_pingBuffer = 0;                // ping-pong style double buffering keeps switching the roles of input and output
float *d_pongBuffer = 0;

CPUBitmap *g_bitmap = 0;


/******************************************************\
 *  GPGPU implementation                              *
\******************************************************/

__global__ void time_step(float *input, float *output, int width, int height, float mu)
{
	int y = blockIdx.y * blockDim.y + threadIdx.y;         // coordinates within 2d array
	int x = blockIdx.x * blockDim.x + threadIdx.x;
    int elemID = y*width + x;                              // index within linear array

	// compute cells needed for update (neighbors + central element)
    // boolean values enable border handling without thread divergence
	int borderFlag = (x > 0);
	float leftNeighb = input[elemID - borderFlag];

    borderFlag = (x < (width - 1));
	float rightNeighb = input[elemID + borderFlag];

    // unary minus turns boolean value into boolean bitwise mask
	borderFlag = -(y > 0);
	float topNeighb = input[elemID - (borderFlag & width)];	

    borderFlag = -(y < (height - 1));
	float bottomNeighb = input[elemID + (borderFlag & width)];
    
	float currElement = input[elemID];
	
	output[elemID] = currElement +
        mu * ((leftNeighb-currElement) +
              (rightNeighb-currElement) +
              (bottomNeighb-currElement) +
              (topNeighb-currElement));
}


/******************************************\
 *  functions used during GLUT callbacks  *
\******************************************/


// compute image
void computeSimulation(int imgWidth, int imgHeight, float *d_pingBuffer, float *d_pongBuffer, float *h_Buffer)
{
    // for measurement of runtimes
	CudaTimer timer;
	float elapsedTime = 0.0f;     

	
	// call CUDA kernel to run on the GPU, copy results back to host, and measure required time
	// set threads layout;  we utilize the fact that img_size = k*32 (see above)
	// consequence: blocks.x * threads.x = img_size
	// -> we don't need to check in the kernel whether the thread is outside the image!
    dim3 cudaBlockSize(16,16);  // image is subdivided into rectangular tiles for parallelism - this variable controls tile size
    dim3 cudaGridSize(imgWidth/cudaBlockSize.x, imgHeight/cudaBlockSize.y);
	timer.start();
	auto buf1 = d_pingBuffer;
	auto buf2 = d_pongBuffer;
	auto tmp = buf1;
	for (int i = 50; i > 0; i--)  // do several simulation steps per displayed frame
	{
		// we do time steps in pairs, so that we don't need to be aware of buffer ping pong everywhere
		for (int j = 2; j > 0; j--) { // do several simulation steps per displayed frame
			time_step<<<cudaGridSize, cudaBlockSize>>>(buf1, buf2, imgWidth, imgHeight, g_mu);
			tmp = buf1;
			buf1 = buf2;
			buf2 = tmp;
		}
	}
	cudaDeviceSynchronize();  // wait until GPU is finished
	handleCUDAerror(__LINE__);

	// copy GPU result back to host memory
    cudaMemcpy(h_Buffer, d_pingBuffer, imgWidth*imgHeight*sizeof(float), cudaMemcpyDeviceToHost);
	handleCUDAerror(__LINE__);
	elapsedTime = timer.elapsed();
	printf("calculation took %4.2f milliseconds.\r", elapsedTime);
}

void cleanUp()
{
	// free memory buffers
    cudaFree(d_pingBuffer);
    cudaFree(d_pongBuffer);
    cudaFreeHost(h_Buffer);
	handleCUDAerror(__LINE__);
    printf("Memory deallocated successfully\n");
}

// static method used for glut callbacks
void keyboard(unsigned char key, int x, int y)
{
	switch (key) 
	{
	case 27:              // keycode for the Escape key
		cleanUp();
		exit(0);
		break;
	}

	glutPostRedisplay();
}

// calls computation, pixel conversion, and OpenGL draw when GLUT refreshes the display
void display( void ) 
{
	computeSimulation(g_imgWidth, g_imgHeight, d_pingBuffer, d_pongBuffer, h_Buffer);
	ue_convertHeat2Bitmap(h_Buffer, g_imgWidth, g_imgHeight, g_bitmap);
	g_bitmap->draw();
	glFlush();
}

// keep the simulation running as fast as possible with a GLUT idle callback
void idleFunc( void )
{
	glutPostRedisplay();
}

/************************
 *  Setup and Test Rig  *
\************************/

int main(int argc, char **argv)
{
	// initialize GLUT, open display window, and register callback functions
	glutInit( &argc, argv );
	glutInitDisplayMode( GLUT_SINGLE | GLUT_RGBA );
	glutInitWindowSize( g_imgWidth, g_imgHeight );
	glutCreateWindow( "Heat Flow - press ESC to close" );
	glutKeyboardFunc(keyboard);
	glutDisplayFunc(display);
	glutIdleFunc(idleFunc);

	// setup CUDA
	ue_checkGPU();
	
	// allocate buffer(s) in main memory and on the GPU
    cudaMalloc(&d_pingBuffer, g_imgWidth*g_imgHeight*sizeof(float));
    cudaMalloc(&d_pongBuffer, g_imgWidth*g_imgHeight*sizeof(float));
    cudaMallocHost(&h_Buffer, g_imgWidth*g_imgHeight*sizeof(float));
	handleCUDAerror(__LINE__);     // in principle, every single CUDA call should be followed by an
	                               // error check. we skipped some to reduce visual clutter. you can
								   // always add the checks to hunt down bugs.
	g_bitmap = new CPUBitmap(g_imgWidth, g_imgHeight);
    printf("Memory allocated successfully\n");

	// initialize the simulation with some starting conditions
	for (int y = g_imgHeight - 1; y >= 0; y--)
	{
		for (int x = g_imgWidth - 1; x >= 0; x--)
		{
            // position for source
			int xOffset = x - g_imgWidth/3;
			int yOffset = y - g_imgHeight/3;

			if (xOffset*xOffset + yOffset*yOffset < 2000)
			{
				h_Buffer[y*g_imgWidth + x] = 40.0f;            // one hot spot
			}
			else
			{
                // position for source
				xOffset = x - 2*g_imgWidth/3;
				yOffset = y - 2*g_imgHeight/3;
				if (xOffset*xOffset + yOffset*yOffset < 2000)
				{
					h_Buffer[y*g_imgWidth + x] = -10.0f;       // one cold spot
				}
				else
				{
                    // position for source
					xOffset = x - g_imgWidth/3;
					yOffset = y - 2*g_imgHeight/3;
					if (xOffset*xOffset + yOffset*yOffset < 200)
					{
						h_Buffer[y*g_imgWidth + x] = 3000.0f;    // a searingly hot spot
					}
					else
					{
						h_Buffer[y*g_imgWidth + x] = 0.f;    // on a neutral background
					}
				}
			}
		}
	}

	cudaMemcpy(d_pingBuffer, h_Buffer, g_imgWidth*g_imgHeight*sizeof(float), cudaMemcpyHostToDevice);
	handleCUDAerror(__LINE__);


	glutMainLoop();

	exit(EXIT_SUCCESS);
}
